"""
Dataset generators.
"""
import numpy as np
from scipy import stats


class NormalGenerator:
    """
    Generate data, can be LRs or scores, for two independent normal distributions: H1 and H2.
    """

    def __init__(self, mu0, sigma0, mu1, sigma1):
        """
        Initializes the generator with two normal distributions.

        :param mu0: mean of class 0 scores (H2)
        :param sigma0: standard deviation of class 0 scores (H2)
        :param mu1: mean of class 1 scores (H1)
        :param sigma1: standard deviation of class 1 scores (H1)
        """
        self.mu0 = mu0
        self.sigma0 = sigma0
        self.mu1 = mu1
        self.sigma1 = sigma1

    def sample_lrs(self, n0, n1):
        """
        Samples LRs in the form P(x|H1)/P(x|H2) for samples from the H2 distribution and the H1 distribution.

        :param n0: number of LRs from class 0 (H2)
        :param n1: number of LRs from class 1 (H1)
        :returns: an array of LRs and an array of labels (value 0 or 1)
        """
        return self._return_probabilities_or_lrs(n0, n1, prob=False)

    def sample_probabilities(self, n0, n1):
        """
        Sample probabilities in the form P(H1) for samples from the H2 distribution and the H1 distribution.

        :param n0: Number of scores from class 0 (H2)
        :param n1: Number of scores from class 1 (H1)
        :return: an array of scores and an array of labels. The scores represent P(H1|x)
        """
        return self._return_probabilities_or_lrs(n0, n1, prob=True)

    def _return_probabilities_or_lrs(self, n0, n1, prob=True):
        """
        Sample probabilities  from both distributions and return either the probabilities  or the o LRs.
        """
        X = np.concatenate([np.random.normal(loc=self.mu0, scale=self.sigma0, size=n0),
                            np.random.normal(loc=self.mu1, scale=self.sigma1, size=n1)])

        p0 = stats.norm.pdf(X, self.mu0, self.sigma0)
        p1 = stats.norm.pdf(X, self.mu1, self.sigma1)

        y = np.concatenate([np.zeros(n0), np.ones(n1)])

        odds = p1 / p0

        return (odds / (1+odds), y) if prob else (odds, y)


class RandomFlipper:
    """
    Random mutilation of a dataset.

    TODO: this class is broken
    """
    def __init__(self, base_generator, p):
        self.gen = base_generator
        self.p = p

    def sample_lrs(self, n0, n1):
        lr, y = self.gen.sample_lrs(n0, n1)
        y[np.random.randint(0, len(y), int(self.p*n0))] = 0
        y[np.random.randint(0, len(y), int(self.p*n1))] = 1

        return lr, y



import numpy as np
from sklearn.isotonic import IsotonicRegression
from sklearn.utils import check_array, check_consistent_length


class IsotonicRegressionInf(IsotonicRegression):
    """
    Sklearn implementation IsotonicRegression throws an error when values are Inf or -Inf when in fact IsotonicRegression
    can handle infinite values. This wrapper around the sklearn implementation of IsotonicRegression prevents the error
    being thrown when Inf or -Inf values are provided.
    """
    def fit(self, X, y, sample_weight=None):
        """Fit the model using X, y as training data.

        Parameters
        ----------
        X : array-like of shape (n_samples,)
            Training data.

        y : array-like of shape (n_samples,)
            Training target.

        sample_weight : array-like of shape (n_samples,), default=None
            Weights. If set to None, all weights will be set to 1 (equal
            weights).

        Returns
        -------
        self : object
            Returns an instance of self.

        Notes
        -----
        X is stored for future use, as :meth:`transform` needs X to interpolate
        new input data.
        """
        check_params = dict(accept_sparse=False, ensure_2d=False, force_all_finite=False)
        X = check_array(X, dtype=[np.float64, np.float32], **check_params)
        y = check_array(y, dtype=X.dtype, **check_params)
        check_consistent_length(X, y, sample_weight)

        # Transform y by running the isotonic regression algorithm and
        # transform X accordingly.
        X, y = self._build_y(X, y, sample_weight)

        # It is necessary to store the non-redundant part of the training set
        # on the model to make it possible to support model persistence via
        # the pickle module as the object built by scipy.interp1d is not
        # picklable directly.
        self._necessary_X_, self._necessary_y_ = X, y

        # Build the interpolation function
        self._build_f(X, y)
        return self

    def transform(self, T):
        """Transform new data by linear interpolation

        Parameters
        ----------
        T : array-like of shape (n_samples,)
            Data to transform.

        Returns
        -------
        T_ : array, shape=(n_samples,)
            The transformed data
        """

        if hasattr(self, '_necessary_X_'):
            dtype = self._necessary_X_.dtype
        else:
            dtype = np.float64

        T = check_array(T, dtype=dtype, ensure_2d=False, force_all_finite=False)

        if len(T.shape) != 1:
            raise ValueError("Isotonic regression input should be a 1d array")

        # Handle the out_of_bounds argument by clipping if needed
        if self.out_of_bounds not in ["raise", "nan", "clip"]:
            raise ValueError("The argument ``out_of_bounds`` must be in "
                             "'nan', 'clip', 'raise'; got {0}"
                             .format(self.out_of_bounds))

        if self.out_of_bounds == "clip":
            T = np.clip(T, self.X_min_, self.X_max_)

        res = self.f_(T)

        # on scipy 0.17, interp1d up-casts to float64, so we cast back
        res = res.astype(T.dtype)

        return res



import numpy as np

def negative_log_likelihood_balanced(X, y, model, params):
    """
    calculates neg_llh of probabilistic binary classifier.
    The llh is balanced in the sense that the total weight of '1'-labels is equal to the total weight of '0'-labels.

    inputs:
        X: n * 1 np.array of scores
        y: n * 1 np.array of labels (Booleans). H1 --> 1, H2 --> 0.
        model: model that links score to posterior probabilities
        params: mapping of parameter names to values of the model.
    output:
        neg_llh_balanced: float, balanced negative log likelihood (base = exp)
    """

    probs = model(X, *params)
    neg_llh_balanced = -np.sum(np.log(probs ** y * (1 - probs) ** (1 - y)) / (y * np.sum(y) + (1 - y) * np.sum(1 - y)))
    return neg_llh_balanced


"""
A multiclass calibrator implements at least the following two methods:

fit(self, X, y)

    X is a two dimensional array of scores; rows are samples; columns are scores 0..1 per class
    y is a one dimensional array of classes 0..n
    returns self

transform(self, X)

    X is a two dimensional array of scores, as in fit()
    returns a two dimensional array of lrs; same dimensions as X
"""
import logging
import math
import warnings

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from sklearn.linear_model import LogisticRegression
from sklearn.neighbors import KernelDensity

from ..calibration import DummyCalibrator
from .util import to_odds, get_classes_from_scores_Xy, to_probability

LOG = logging.getLogger(__name__)


class LogitCalibrator(BaseEstimator, TransformerMixin):
    """
    Calculates a likelihood ratio of a score value, provided it is from one of
    several distributions. Uses logistic regression for interpolation.
    """

    def fit(self, X, y):
        self._logit = LogisticRegression(class_weight='balanced')
        self._logit.fit(X, y)

        return self

    def transform(self, X):
        self.p = self._logit.predict_proba(X)
        lrs = to_odds(self.p)
        return lrs


class BalancedPriorCalibrator(BaseEstimator, TransformerMixin):
    """
    Recalculates LRs as posterior odds for balanced priors.

    In a forensic context, an LR of two classes is the same as posterior odds
    if the priors for both classes are equal. In multiclass classification,
    classifiers tend to produce posterior probabilities based on balanced
    classes, i.e. every class has a prior probability of 1/n, with n is the
    number of classes. This leads to prior odds of every class versus all other
    classes of (1/n) / (n-1)/n = 1/(n-1).

    This class compensates for that by updating the prior odds to 1 and update
    posterior odds (=LRs) accordingly.
    """
    def __init__(self, backend):
        self.backend = backend

    def fit(self, X, y):
        self.backend.fit(X, y)
        return self

    def transform(self, X):
        X = to_probability(self.backend.transform(X))
        self.priors = np.ones(X.shape[1]) / X.shape[1]

        priors_sum = np.sum(self.priors)
        prior_odds = self.priors / (priors_sum - self.priors)
        lrs = to_odds(X) / prior_odds
        self.p = to_probability(lrs)
        return lrs



import logging

import numpy as np
import sklearn
import sklearn.mixture
from sklearn.pipeline import Pipeline

from .metrics import calculate_lr_statistics, LrStats
from .transformers import EstimatorTransformer, ComparisonFunctionTransformer

from .util import Xn_to_Xy, LR

LOG = logging.getLogger(__name__)


def _create_transformer(scorer):
    if hasattr(scorer, "transform"):
        return scorer
    elif hasattr(scorer, "predict_proba"):
        return EstimatorTransformer(scorer)
    elif callable(scorer):
        return ComparisonFunctionTransformer(scorer)
    else:
        raise NotImplementedError("`scorer` argument must either be callable or implement at least one of `transform`, `predict_proba`")


class CalibratedScorer:
    """
    LR system which produces calibrated LRs using a scorer and a calibrator.

    The scorer is an object that transforms instance features into a scalar value (a "score"). The scorer may either
    be:
     - an estimator (e.g. `sklearn.linear_model.LogisticRegression`) object which implements `fit` and `predict_proba`;
     - a transformer object which implements `transform` and optionally `fit`; or
     - a distance function (callable) that takes paired instances as its arguments and returns a distance for each pair.

    The scorer can also be a composite object such as a `sklearn.pipeline.Pipeline`. If the scorer is an estimator, the
    probabilities it produces are transformed to their log odds.

    The calibrator is an object that transforms instance scores to LRs.

    This class supports `fit`, which fits the scorer and the calibrator on the same data. If only the calibrator is to
    be fit, use `fit_calibrator`. Both the scorer and the calibrator can be accessed by their attributes `scorer` and
    `calibrator`.
    """
    def __init__(self, scorer, calibrator):
        """
        Constructor.

        Parameters
        ----------
        scorer an object that transforms features into scores.
        calibrator an object that transforms scores into LRs.
        """
        self.scorer = _create_transformer(scorer)
        self.calibrator = calibrator
        self.pipeline = Pipeline([
            ("scorer", self.scorer),
            ("reshape", sklearn.preprocessing.FunctionTransformer(self._reshape)),
            ("calibrator", self.calibrator)
        ])

    @staticmethod
    def _reshape(X):
        if len(X.shape) == 1:
            return X
        else:
            assert len(X) == X.shape[0], f"array has bad dimensions: all dimensions but the first should be 1; found {X.shape}"
            return X.reshape(-1)

    def fit(self, X, y):
        """
        Fits both the scorer and the calibrator on the same data.

        Parameters
        ----------
        X the feature vectors of the instances
        y the instance labels

        Returns
        -------
        `self`
        """
        self.pipeline.fit(X, y)
        return self

    def fit_calibrator(self, X, y):
        """
        Fits the calibrator without modifying the scorer.

        The arguments are the same as in `fit`. Before calibrating, this method transforms the feature vectors into
        scores by calling `transform` on the scorer.

        Parameters
        ----------
        X the feature vectors of the instances
        y the instance labels

        Returns
        -------
        `self`
        """
        X = self.scorer.transform(X)
        X = self._reshape(X)
        self.calibrator.fit(X, y)
        return self

    def predict_lr(self, X):
        """
        Compute LRs for instances.

        Parameters
        ----------
        X the feature vector of the instances

        Returns
        -------
        a vector of LRs
        """
        return self.pipeline.transform(X)


class CalibratedScorerCV:
    """
    Variant of `CalibratedScorer` that uses cross-calibration to optimize for small data sets.
    """
    def __init__(self, scorer, calibrator, n_splits):
        self.scorer = _create_transformer(scorer)
        self.calibrator = calibrator
        self.n_splits = n_splits

    def fit(self, X, y):
        kf = sklearn.model_selection.StratifiedKFold(n_splits=self.n_splits, shuffle=True)

        Xcal = np.empty([0, 1])
        ycal = np.empty([0])
        for train_index, cal_index in kf.split(X, y):
            self.scorer.fit(X[train_index], y[train_index])
            p = self.scorer.transform(X[cal_index])
            Xcal = np.append(Xcal, p)
            ycal = np.append(ycal, y[cal_index])
        self.calibrator.fit(Xcal, ycal)

        self.scorer.fit(X, y)
        return self

    def predict_lr(self, X):
        scores = self.scorer.transform(X)
        return self.calibrator.transform(scores)


def calibrate_lr(point, calibrator):
    """
    Calculates a calibrated likelihood ratio (LR).

    P(E|H1) / P(E|H0)
    
    Parameters
    ----------
    point : float
        The point of interest for which the LR is calculated. This point is
        sampled from either class 0 or class 1.
    calibrator : a calibration function
        A calibration function which is fitted for two distributions.

    Returns
    -------
    tuple<3> of float
        A likelihood ratio, with positive values in favor of class 1, and the
        probabilities P(E|H0), P(E|H1).
    """
    lr = calibrator.transform(np.array(point))[0]
    return LR(lr, calibrator.p0[0], calibrator.p1[0])


def apply_scorer(scorer, X):
    return scorer.predict_proba(X)[:,1] # probability of class 1


def scorebased_lr(scorer, calibrator, X0_train, X1_train, X0_calibrate, X1_calibrate, X_disputed):
    """
    Trains a classifier, calibrates the outcome and calculates a LR for a single
    sample.

    P(E|H1) / P(E|H0)

    All numpy arrays have two dimensions, representing samples and features.

    Parameters
    ----------
    scorer : classifier
        A model to be trained. Must support probability output.
    X0_train : numpy array
        Training set of samples from class 0
    X1_train : numpy array
        Training set of samples from class 1
    X0_calibrate : numpy array
        Calibration set of samples from class 0
    X1_calibrate : numpy array
        Calibration set of samlpes from class 1
    X_disputed : numpy array
        Test samples of unknown class

    Returns
    -------
    float
        A likelihood ratio for `X_disputed`
    """
    Xtrain, ytrain = Xn_to_Xy(X0_train, X1_train)
    scorer.fit(Xtrain, ytrain)

    Xcal, ycal = Xn_to_Xy(X0_calibrate, X1_calibrate)
    calibrator.fit(scorer.predict_proba(Xcal)[:,1], ycal)
    scorer = CalibratedScorer(scorer, calibrator)

    return scorer.predict_lr(X_disputed)


def calibrated_cllr(calibrator, class0_calibrate, class1_calibrate, class0_test=None, class1_test=None) -> LrStats:
    Xcal, ycal = Xn_to_Xy(class0_calibrate, class1_calibrate)
    calibrator.fit(Xcal, ycal)

    use_calibration_set_for_test = class0_test is None
    if use_calibration_set_for_test:
        class0_test = class0_calibrate
        class1_test = class1_calibrate

    lrs0 = calibrator.transform(class0_test)
    lrs1 = calibrator.transform(class1_test)

    lrs0 = [ LR(*stats) for stats in zip(lrs0, calibrator.p0, calibrator.p1) ]
    lrs1 = [ LR(*stats) for stats in zip(lrs1, calibrator.p0, calibrator.p1) ]

    return calculate_lr_statistics(lrs0, lrs1)


def scorebased_cllr(scorer, calibrator, X0_train, X1_train, X0_calibrate, X1_calibrate, X0_test=None, X1_test=None) -> LrStats:
    """
    Trains a classifier on a training set, calibrates the outcome with a
    calibration set, and calculates a LR (likelihood ratio) for all samples in
    a test set. Calculates a Cllr (log likelihood ratio cost) value for the LR
    values. If no test set is provided, the calibration set is used for both
    calibration and testing.

    All numpy arrays have two dimensions, representing samples and features.

    Parameters
    ----------
    scorer : classifier
        A model to be trained. Must support probability output.
    X0_train : numpy array
        Training set for class 0
    X1_train : numpy array
        Training set for class 1
    X0_calibrate : numpy array
        Calibration set for class 0
    X1_calibrate : numpy array
        Calibration set for class 1
    X0_test : numpy array
        Test set for class 0
    X1_test : numpy array
        Test set for class 1

    Returns
    -------
    float
        A likelihood ratio for `X_test`
    """
    if X0_test is None:
        LOG.debug('scorebased_cllr: training_size: {train0}/{train1}; calibration size: {cal0}/{cal1}'.format(train0=X0_train.shape[0], train1=X1_train.shape[0], cal0=X0_calibrate.shape[0], cal1=X1_calibrate.shape[0]))
    else:
        LOG.debug('scorebased_cllr: training_size: {train0}/{train1}; calibration size: {cal0}/{cal1}; test size: {test0}/{test1}'.format(train0=X0_train.shape[0], train1=X1_train.shape[0], cal0=X0_calibrate.shape[0], cal1=X1_calibrate.shape[0], test0=X0_test.shape[0], test1=X1_test.shape[0]))

    scorer.fit(*Xn_to_Xy(X0_train, X1_train))

    class0_test = apply_scorer(scorer, X0_test) if X0_test is not None else None
    class1_test = apply_scorer(scorer, X1_test) if X0_test is not None else None

    return calibrated_cllr(calibrator, apply_scorer(scorer, X0_calibrate), apply_scorer(scorer, X1_calibrate), class0_test, class1_test)


def scorebased_lr_kfold(scorer, calibrator, n_splits, X0_train, X1_train, X_disputed):
    scorer = CalibratedScorerCV(scorer, calibrator, n_splits=n_splits)
    scorer.fit(*Xn_to_Xy(X0_train, X1_train))
    return scorer.predict_lr(X_disputed)


def scorebased_cllr_kfold(scorer, calibrator, n_splits, X0_train, X1_train, X0_test, X1_test) -> LrStats:
    LOG.debug('scorebased_cllr_kfold: training_size: {train0}/{train1}; test size: {test0}/{test1}'.format(train0=X0_train.shape[0], train1=X1_train.shape[0], test0=X0_test.shape[0], test1=X1_test.shape[0]))

    X_disputed = np.concatenate([X0_test, X1_test])
    lrs = scorebased_lr_kfold(scorer, calibrator, n_splits, X0_train, X1_train, X_disputed)

    assert X0_test.shape[0] + X1_test.shape[0] == len(lrs)
    lrs0 = lrs[:X0_test.shape[0]]
    lrs1 = lrs[X0_test.shape[0]:]

    lrs0 = [ LR(*stats) for stats in zip(lrs0, calibrator.p0, calibrator.p1) ]
    lrs1 = [ LR(*stats) for stats in zip(lrs1, calibrator.p0, calibrator.p1) ]

    return calculate_lr_statistics(lrs0, lrs1)



from typing import Callable, Optional

import numpy as np
import sklearn
from scipy.interpolate import interp1d
from scipy.stats import rankdata
import warnings

from sklearn.base import TransformerMixin
from sklearn.preprocessing import FunctionTransformer

from lir.util import to_log_odds


class EstimatorTransformer(TransformerMixin):
    """
    A wrapper for an estimator to make it behave like a transformer.

    In particular, it implements `transform` by calling `predict_proba` on the underlying estimator, and transforming
    the probabilities to their corresponding log odds value. Optionally, an alternative transformation function can be
    specified.
    """
    def __init__(self, estimator, transform_probabilities: Optional[Callable] = to_log_odds):
        self.estimator = estimator
        self.transform_probabilities = transform_probabilities

    def fit(self, X, y):
        self.estimator.fit(X, y)
        return self

    def transform(self, X):
        return self.transform_probabilities(self.estimator.predict_proba(X)[:, 1])

    def __getattr__(self, item):
        return getattr(self.estimator, item)


class ComparisonFunctionTransformer(FunctionTransformer):
    """
    A wrapper for a distance function to make it behave like a transformer.

    This is essentially a `FunctionTransformer` except that it expects a function takes two arguments of the same
    dimensions: `X` and `Y`, where each row in `X` is compared to the row with the same index in `Y`.

    See: the `paired_*` distance functions in `sklearn.metrics.pairwise`.
    """
    def __init__(self, distance_function: Callable):
        self.distance_function = distance_function
        super().__init__(self.transformer_function)

    def transformer_function(self, X):
        if len(X.shape) == 2:
            return self.distance_function(X)
        elif len(X.shape) == 3:
            vectors = [X[:, :, i] for i in range(X.shape[2])]
            return self.distance_function(*vectors)
        else:
            raise ValueError(f"unexpected input shape; expected: (*, *) or (*, *, *); found: {X.shape}")


class AbsDiffTransformer(sklearn.base.TransformerMixin):
    """
    Takes an array of sample pairs and returns the element-wise absolute difference.

    Expects:
        - X is of shape (n,f,2) with n=number of pairs; f=number of features; 2=number of samples per pair;
    Returns:
        - X has shape (n, f)
    """
    def fit(self, X, y=None):
        return self

    def transform(self, X):
        assert len(X.shape) == 3
        assert X.shape[2] == 2

        return np.abs(X[:,:,0] - X[:,:,1])


class PercentileRankTransformer(sklearn.base.TransformerMixin):
    """
    Compute the percentile rankings of a dataset, relative to another dataset.
    Rankings are in range [0, 1]. Handling ties: the maximum of the ranks that
    would have been assigned to all the tied values is assigned to each value.

    To be able to compute the rankings of dataset Z relative to dataset X,
    'fit' will create a ranking function for each feature, based on X.
    'transform' will apply ranking of Z based on dataset X.

    Fit:
    Expects:
        - X is of shape (n, f) with n = number of measurements,
        f = number of features

    Transform:
    Expects:
        - X is of shape (m, f) with m = number of measurements,
        f = number of features
    Returns:
        - rankings with shape (m, f)
    """
    def __init__(self):
        self.rank_functions = None

    def fit(self, X, y=None):
        assert len(X.shape) == 2
        ranks_X = rankdata(X, method='max', axis=0)/len(X)
        # if a feature is a constant int value, interp1d returns nan -> convert to float
        X = X.astype(float)
        self.rank_functions = [interp1d(X[:, i], ranks_X[:, i], bounds_error=False,
                                        fill_value=(0, 1)) for i in range(X.shape[1])]
        return self

    def transform(self, X):
        assert self.rank_functions, "transform() called before fit()"
        assert len(X.shape) == 2
        assert X.shape[1] == len(self.rank_functions),\
            "number of features used for fit() and transform() should be equal"
        ranks = [self.rank_functions[i](X[:, i]) for i in range(X.shape[1])]
        return np.stack(ranks, axis=1)


class InstancePairing(sklearn.base.TransformerMixin):
    def __init__(self,
                 same_source_limit=None,
                 different_source_limit=None,
                 ratio_limit=None,
                 seed=None):
        """
        Creates pairs of instances.

        This transformer takes `X` and `y` as input, and returns `X_paired` and `y_paired`.
        `X_paired` is the carthesian product of `X` and itself.
        In other words, `X_paired` contains all possible pairs of feature vectors in `X`, meaning that
        `X_paired.shape[0] == X.shape[0]*X.shape[0]`, except if their number exceeds the values of parameters
        `same_source_limit` and `different_source_limit`, in which case pairs are randomly drawn from the full set of
        pairs. Also, a maximum ratio between the same and different source pairs can be specified with 'ratio_limit'.

        Note that this transformer may cause performance problems with large datasets,
        even if the number of instances in the output is limited.

        Parameters:
            - same_source_limit (int or None): the maximum number of same source pairs (None = no limit)
            - different_source_limit (int or None or 'balanced'): the maximum number of different source pairs (None = no limit; 'balanced' = number of same source pairs)
            - ratio_limit (int or None): maximum ratio between same source and different source pairs.
                Ratio = ds pairs / ss pairs. The number of ds pairs will not exceed ratio_limit * ss pairs.
                If both ratio and same_source_limit/different_source_limit are specified,
                the number of pairs is chosen such that the ratio_limit is preserved and
                the limit(s) are not exceeded, while taking as many pairs as possible within these constraints.
            - seed (int or None): seed to make pairing reproducible
        """
        self._ss_limit = same_source_limit
        self._ds_limit = different_source_limit
        self._ratio_limit = ratio_limit
        self.rng = np.random.default_rng(seed=seed)

        if self._ds_limit == 'balanced':
            warnings.warn('The argument \'balanced\' is deprecated. '
                          'Use ratio_limit instead.', DeprecationWarning, stacklevel=2)
            self._ds_limit = None
            self._ratio_limit = 1

    def fit(self, X):
        return self

    def transform(self, X, y):
        """
        Expects:
            - X is of shape (n,f) with n=number of samples; f=number of features
            - y is of shape (n,)
        Returns:
            - X_paired, y_paired: X_paired has shape (m,f,2); y_paired has shape (m,); with m=number of pairs.
                For example, `(X_paired[a,:,0], X_paired[a,:,1])` is the a^th pair of feature vectors, and `y_paired[a]`
                is 1 if the labels of both vectors in `y` are equal (same source), or 0 otherwise (different source).
        Attributes:
            - self.pairing: an array of shape (m,2) with indexes of X input array which contribute to the pairs
        """
        pairing = np.array(np.meshgrid(np.arange(X.shape[0]), np.arange(X.shape[0]))).T.reshape(-1, 2)  # generate all possible pairs
        same_source = y[pairing[:, 0]] == y[pairing[:, 1]]

        rows_same = np.where((pairing[:, 0] < pairing[:, 1]) & same_source)[0]  # pairs with different id and same source
        rows_diff = np.where((pairing[:, 0] < pairing[:, 1]) & ~same_source)[0]  # pairs with different id and different source

        if self._ss_limit is not None and rows_same.size > self._ss_limit:
            rows_same = self.rng.choice(rows_same, self._ss_limit, replace=False)

        n_ds_pairs = min(x for x in [rows_same.size * self._ratio_limit if self._ratio_limit else None,
                                     self._ds_limit,
                                     rows_diff.size
                                     ] if x is not None)

        if n_ds_pairs < rows_diff.size:
            rows_diff = self.rng.choice(rows_diff, n_ds_pairs, replace=False)

        pairing = np.concatenate([pairing[rows_same, :], pairing[rows_diff, :]])
        self.pairing = pairing

        X = np.stack([X[pairing[:, 0]], X[pairing[:, 1]]], axis=2)  # pair instances by adding another dimension
        y = np.concatenate([np.ones(rows_same.size), np.zeros(rows_diff.size)])  # apply the new labels: 1=same_source versus 0=different_source

        return X, y



import collections
import warnings
from typing import List

import numpy as np

from .calibration import IsotonicCalibrator
from .util import Xn_to_Xy, Xy_to_Xn, to_probability, LR


LrStats = collections.namedtuple('LrStats',
                                 ['avg_log2lr', 'avg_log2lr_class0', 'avg_log2lr_class1', 'avg_p0_class0', 'avg_p1_class0',
                                  'avg_p0_class1', 'avg_p1_class1', 'cllr_class0', 'cllr_class1', 'cllr', 'lr_class0',
                                  'lr_class1', 'cllr_min', 'cllr_cal'])


def cllr(lrs, y, weights=(1, 1)):
    """
    Calculates a log likelihood ratio cost (C_llr) for a series of likelihood
    ratios.

    Nico Brümmer and Johan du Preez, Application-independent evaluation of speaker detection, In: Computer Speech and
    Language 20(2-3), 2006.

    Parameters
    ----------
    lrs : a numpy array of LRs
    y : a numpy array of labels (0 or 1)

    Returns
    -------
    cllr
        the log likelihood ratio cost
    """

    # ignore errors:
    #   divide -> ignore divide by zero
    #   over -> ignore scalar overflow
    with np.errstate(divide='ignore', over='ignore'):
        lrs0, lrs1 = Xy_to_Xn(lrs, y)
        cllr0 = weights[0] * np.mean(np.log2(1 + lrs0)) if weights[0] > 0 else 0
        cllr1 = weights[1] * np.mean(np.log2(1 + 1 / lrs1)) if weights[1] > 0 else 0
        return (cllr0 + cllr1) / sum(weights)


def cllr_min(lrs, y, weights=(1, 1)):
    """
    Estimates the discriminative power from a collection of likelihood ratios.

    Parameters
    ----------
    lrs : a numpy array of LRs
    y : a numpy array of labels (0 or 1)

    Returns
    -------
    cllr_min
        the log likelihood ratio cost
    """
    cal = IsotonicCalibrator()
    lrmin = cal.fit_transform(to_probability(lrs), y)
    return cllr(lrmin, y, weights)


def devpav_estimated(lrs, y, resolution=1000):
    """
    Estimate devPAV, a metric for calibration.

    devPAV is the cumulative deviation of the PAV transformation from
    the identity line. It is calculated in the LR range where misleading LRs
    occur.

    See also: P. Vergeer, Measuring calibration of likelihood ratio systems: a
    comparison of four systems, including a new metric devPAV, to appear

    This implementation estimates devPAV by calculating the average deviation
    for a large number of LRs.

    Parameters
    ----------
    lrs : a numpy array of LRs
    y : a numpy array of labels (0 or 1)
    resolution : the number of measurements in the range of misleading evidence; a higher value yields a more accurate estimation

    Returns
    -------
    devPAV
        an estimation of devPAV
    """
    lrs0, lrs1 = Xy_to_Xn(lrs, y)
    if len(lrs0) == 0 or len(lrs1) == 0:
        raise ValueError('devpav: illegal input: at least one value is required for each class')

    # find misleading LR extremes
    first_misleading = np.min(lrs1)
    last_misleading = np.max(lrs0)
    if first_misleading > last_misleading:  # test for perfect discrimination
        return 0

    if np.isinf(first_misleading) or np.isinf(last_misleading):  # test for infinitely misleading LRs
        return np.inf

    # calibrate on the input LRs
    cal = IsotonicCalibrator()
    cal.fit_transform(to_probability(lrs), y)

    # take `resolution` points evenly divided along the range of misleading LRs
    xlr = np.exp(np.linspace(np.log(first_misleading), np.log(last_misleading), resolution))
    pavlr = cal.transform(to_probability(xlr))

    devlr = np.absolute(np.log10(xlr) - np.log10(pavlr))
    return (np.sum(devlr) / resolution) * (np.log10(last_misleading) - np.log10(first_misleading))


def calcsurface_f(c1, c2):
    """
    Helperfunction that calculates the desired surface for two xy-coordinates
    """
    # step 1: calculate intersection (xs, ys) of straight line through coordinates with identity line (if slope (a) = 1, there is no intersection and surface of this parrellogram is equal to deltaY * deltaX)


    x1, y1 = c1
    x2, y2 = c2
    a = (y2 - y1) / (x2 - x1)

    if a == 1:
        # dan xs equals +/- Infinite en is er there is no intersection with the identity line
        # since condition 1 holds the product below is always positive
        surface = (y2 - y1) * (x2 - x1)
    elif (a < 0):
        raise ValueError(f"slope is negative; impossible for PAV-transform. Coordinates are {c1} and {c2}. Calculated slope is {a}")
    else:
        # than xs is finite:
        b = y1 - a * x1
        xs = b / (1 - a)
        # xs

        # step 2: check if intersection is located within line segment c1 and c2.
        if x1 < xs and x2 >= xs:
            # then intersection is within
            # (situation 1 of 2) if y1 <= x1 than surface is:
            if y1 <= x1:
                surface = 0.5 * (xs - y1) * (xs - x1) - 0.5 * (xs - x1) * (xs - x1) + 0.5 * (y2 - xs) * (x2 - xs) - 0.5 * (
                            x2 - xs) * (x2 - xs)
            else:
                # (situation 2 of 2) than y1 > x1, and surface is:
                surface = 0.5 * (xs - x1) ** 2 - 0.5 * (xs - y1) * (xs - x1) + 0.5 * (x2 - xs) ** 2 - 0.5 * (x2 - xs) * (
                            y2 - xs)
                # dit is the same as 0.5 * (xs - x1) * (xs - y1) - 0.5 * (xs - y1) * (xs - y1) + 0.5 * (y2 - xs) * (x2 - xs) - 0.5 * (y2 - xs) * (y2 - xs) + 0.5 * (y1 - x1) * (y1 - x1) + 0.5 * (x2 - y2) * (x2 -y2)
        else:  # then intersection is not within line segment
            # if (situation 1 of 4) y1 <= x1 AND y2 <= x1, and surface is
            if y1 <= x1 and y2 <= x1:
                surface = 0.5 * (y2 - y1) * (x2 - x1) + (x1 - y2) * (x2 - x1) + 0.5 * (x2 - x1) * (x2 - x1)
            elif y1 > x1:  # (situation 2 of 4) then y1 > x1, and surface is
                surface = 0.5 * (x2 - x1) * (x2 - x1) + (y1 - x2) * (x2 - x1) + 0.5 * (y2 - y1) * (x2 - x1)
            elif y1 <= x1 and y2 > x1:  # (situation 3 of 4). This should be the last possibility.
                surface = 0.5 * (y2 - y1) * (x2 - x1) - 0.5 * (y2 - x1) * (y2 - x1) + 0.5 * (x2 - y2) * (x2 - y2)
            else:
                # situation 4 of 4 : this situation should never appear. There is a fourth sibution as situation 3, but than above the identity line. However, this is impossible by definition of a PAV-transform (y2 > x1).
                raise ValueError(f"unexpected coordinate combination: ({x1}, {y1}) and ({x2}, {y2})")
    return surface


def _devpavcalculator(lrs, pav_lrs, y):
    """
    function that calculates davPAV for a PAVresult for SSLRs and DSLRs  een PAV transformatie de devPAV uitrekent
    Input: Lrs = np.array met LR-waarden. pav_lrs = np.array met uitkomst van PAV-transformatie op lrs. y = np.array met labels (1 voor H1 en 0 voor H2)
    Output: devPAV value

    """
    DSLRs, SSLRs = Xy_to_Xn(lrs,y)
    DSPAVLRs, SSPAVLRs = Xy_to_Xn(pav_lrs, y)
    PAVresult = np.concatenate([SSPAVLRs, DSPAVLRs])
    Xen = np.concatenate([SSLRs, DSLRs])

    # order coordinates based on x's then y's and filtering out identical datapoints
    data = np.unique(np.array([Xen, PAVresult]), axis=1)
    Xen = data[0,:]
    Yen = data[1,:]
    # pathological cases
    # check if min(Xen) = 0 or max(Xen) = Inf. First min(Xen)
    # eerst van drie: als Xen[0] == 0 en Xen[len(Xen)-1] != Inf
    if Xen[0] == 0 and Xen[-1] != np.inf:
        if Yen[0] == 0 and Yen[1] != 0:
            # dan loopt er een lijn in de PAV transform tot {inf, -Inf} evenwijdig aan de lijn y=x
            return (np.absolute(np.log10(Xen[1]) - np.log10(Yen[1])))
        else:
            # dan is Yen[0] finite of Yen[1] gelijk 0 en loopt er ergens een horizontale lijn tot Log(Xen[0]) = -Inf. devPAV wordt oneindig
            return np.inf
        # tweede van drie: als Xen[len(Xen)-1] == Inf en Xen[0] != 0
    elif Xen[0] != 0 and Xen[-1] == np.inf:
        if Yen[len(Yen) - 1] == np.inf and Yen[len(Yen) - 2] != np.inf:
            # dan loopt er een lijn in de PAV transform tot {inf, -Inf} evenwijdig aan de lijn y=x
            return (np.absolute(np.log10(Xen[len(Xen) - 2]) - np.log10(Yen[len(Yen) - 2])))
        else:
            # dan is Yen[len(Yen] finite of Yen[len(Yen-2] gelijk inf en loopt er ergens een horizontale lijn tot Log(Xen[len(Xen)]) = Inf. devPAV wordt oneindig
            return np.inf
        # derde van drie: als Xen[0] = 0 en Xen[len(Xen)-1] == Inf
    elif Xen[0] == 0 and Xen[-1] == np.inf:
        if Yen[len(Yen) - 1] == np.inf and Yen[len(Yen) - 2] != np.inf and Yen[0] == 0 and Yen[1] != 0:
            # dan zijn de lijnen aan beide uiteinden evenwijdig met de lijn Y=X. Het berekenen van devPAV is het gemiddelde van twee coordinaten
            devPAV = ((np.absolute(np.log10(Xen[len(Xen) - 2]) - np.log10(Yen[len(Yen) - 2]))) + np.absolute(
                np.log10(Xen[1]) - np.log10(Yen[1]))) / 2
            return (devPAV)
        else:
            # dan loopt er ergens een horizontale lijn met oneindig bereik is is devPAV Inf
            return np.inf

    else:
        # dan is het geen pathological case met rare X-waarden en kan devPAV berekend worden

        # filtering out -Inf or 0 Y's
        wh = (Yen > 0) & (Yen < np.inf)
        Xen = np.log10(Xen[wh])
        Yen = np.log10(Yen[wh])
        # create an empty list with size (len(Xen))
        devPAVs = [None] * len(Xen)
        # sanity check
        if len(Xen) == 0:
            return np.nan
        elif len(Xen) == 1:
            return (abs(Xen - Yen))
        # than calculate devPAV
        else:
            deltaX = Xen[-1] - Xen[0]
            surface = (0)
            for i in range(1, (len(Xen))):
                surface = surface + calcsurface_f((Xen[i - 1], Yen[i - 1]), (Xen[i], Yen[i]))
                devPAVs[i - 1] = calcsurface_f((Xen[i - 1], Yen[i - 1]), (Xen[i], Yen[i]))
            # return(list(surface/a, PAVresult, Xen, Yen, devPAVs))
            return (surface / deltaX)


def devpav(lrs, y):
    """
    calculates PAV transform of LR data under H1 and H2.
    """
    cal = IsotonicCalibrator()
    pavlrs = cal.fit_transform(to_probability(lrs), y)
    return _devpavcalculator(lrs, pavlrs, y)


def calculate_lr_statistics(lr_class0: List[LR], lr_class1: List[LR]) -> LrStats:
    """
    Calculates various statistics for a collection of likelihood ratios.

    Parameters
    ----------
    lr_class0 : list of float
        Likelihood ratios which are calculated for measurements which are
        sampled from class 0.
    lr_class1 : list of float
        Likelihood ratios which are calculated for measurements which are
        sampled from class 1.

    Returns
    -------
    LrStats
        Likelihood ratio statistics.
    """
    assert len(lr_class0) > 0
    assert len(lr_class1) > 0

    def avg(*args):
        return sum(args) / len(args)

    if type(lr_class0[0]) == LR:
        avg_p0_class0 = avg(*[lr.p0 for lr in lr_class0])
        avg_p1_class0 = avg(*[lr.p1 for lr in lr_class0])
        avg_p0_class1 = avg(*[lr.p0 for lr in lr_class1])
        avg_p1_class1 = avg(*[lr.p1 for lr in lr_class1])
        lr_class0 = np.array([lr.lr for lr in lr_class0])
        lr_class1 = np.array([lr.lr for lr in lr_class1])
    else:
        if type(lr_class0) == list:
            lr_class0 = np.array(lr_class0)
            lr_class1 = np.array(lr_class1)

        avg_p0_class0 = None
        avg_p1_class0 = None
        avg_p0_class1 = None
        avg_p1_class1 = None

    with warnings.catch_warnings():
        try:
            avg_log2lr_class0 = np.mean(np.log2(1 / lr_class0))
            avg_log2lr_class1 = np.mean(np.log2(lr_class1))
            avg_log2lr = avg(avg_log2lr_class0, avg_log2lr_class1)
        except RuntimeWarning:
            # possibly illegal LRs such as 0 or inf
            avg_log2lr_class0 = np.nan
            avg_log2lr_class1 = np.nan
            avg_log2lr = np.nan

    lrs, y = Xn_to_Xy(lr_class0, lr_class1)
    cllr_class0 = cllr(lrs, y, weights=(1, 0))
    cllr_class1 = cllr(lrs, y, weights=(0, 1))
    cllr_ = .5 * (cllr_class0 + cllr_class1)

    cllrmin_class0 = cllr_min(lrs, y, weights=(1, 0))
    cllrmin_class1 = cllr_min(lrs, y, weights=(0, 1))
    cllrmin = .5 * (cllrmin_class0 + cllrmin_class1)

    return LrStats(avg_log2lr, avg_log2lr_class0, avg_log2lr_class1,
                   avg_p0_class0, avg_p1_class0, avg_p0_class1, avg_p1_class1,
                   cllr_class0, cllr_class1, cllr_, lr_class0, lr_class1,
                   cllrmin, cllr_ - cllrmin)



from setuptools import setup, find_packages

with open("README.md") as f:
    long_description = f.read()
setup(
    name="lir",
    version="0.1.24",
    description="scripts for calculating likelihood ratios",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/NetherlandsForensicInstitute/lir",
    author="Netherlands Forensic Institute",
    author_email="fbda@nfi.nl",
    packages=find_packages(),
    setup_requires=["nose"],
    test_suite="nose.collector",
    install_requires=["matplotlib", "numpy", "scipy", "scikit-learn", "tqdm"],
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



import logging


LOG = logging.getLogger(__name__)


class CalibratedScorer:
    def __init__(self, scorer, calibrator):
        self.scorer = scorer
        self.calibrator = calibrator

    def fit(self, X, y):
        self.fit_scorer(X, y)
        self.fit_calibrator(X, y)

    def fit_scorer(self, X, y):
        self.scorer.fit(X, y)

    def fit_calibrator(self, X, y):
        p = self.scorer.predict_proba(X)
        self.calibrator.fit(p, y)

    def predict(self, X):
        return self.scorer.predict(X)

    def predict_lr(self, X):
        X = self.scorer.predict_proba(X)
        return self.calibrator.transform(X)
